/* -------------------------------------------------------------------------- */

#include <zephyr/kernel.h>
#include <zephyr/random/random.h>
#include <zephyr/sys/byteorder.h>

#include <zephyr/bluetooth/bluetooth.h>
#include <bluetooth/services/nus.h>
#include <bluetooth/scan.h>
#include <dm.h>

#define DEVICE_NAME             CONFIG_BT_DEVICE_NAME
#define DEVICE_NAME_LEN         (sizeof(DEVICE_NAME) - 1)

#define COMPANY_CODE            0x0059
#define SUPPORT_DM_CODE         0xFF55AA5A

/* -------------------------------------------------------------------------- */

static void connected_cb(struct bt_conn* conn, uint8_t err);
static void disconnected_cb(struct bt_conn* conn, uint8_t reason);
static int adv_start(void);
static void adv_scanned_cb(struct bt_le_ext_adv* adv, struct bt_le_ext_adv_scanned_info* info);
static int scan_dm_start(void);
static void scan_filter_dm_match(struct bt_scan_device_info* device_info, struct bt_scan_filter_match* filter_match, bool connectable);
static bool filter_data_dm_cb(struct bt_data* data, void* user_data);
static void scan_filter_nus_match(struct bt_scan_device_info* device_info, struct bt_scan_filter_match* filter_match, bool connectable);
void scan_nus_connecting_error(struct bt_scan_device_info* device_info);
void scan_nus_connecting(struct bt_scan_device_info* device_info, struct bt_conn* conn);
static void adv_work_cb(struct k_work* work);

/* -------------------------------------------------------------------------- */

struct adv_mfg_data {
    uint16_t company_code;	    /* Company Identifier Code. */
    uint32_t support_dm_code;   /* To identify the device that supports distance measurement. */
    uint32_t rng_seed;          /* Random seed used for generating hopping patterns. */
} __packed;

static struct adv_mfg_data mfg_data;
static struct bt_le_ext_adv* adv;
static struct bt_le_adv_param* adv_param;
static struct bt_conn* current_connection;

BT_CONN_CB_DEFINE(conn_callbacks) = {
    .connected = connected_cb,
    .disconnected = disconnected_cb,
};

BT_SCAN_CB_INIT(scan_nus_cb, scan_filter_nus_match, NULL, scan_nus_connecting_error, scan_nus_connecting);
BT_SCAN_CB_INIT(scan_dm_cb, scan_filter_dm_match, NULL, NULL, NULL);

static K_WORK_DEFINE(adv_work, adv_work_cb);

static struct bt_le_scan_param scan_param = {
    .type     = BT_LE_SCAN_TYPE_ACTIVE,
    .interval = BT_GAP_SCAN_FAST_INTERVAL,
    .window   = BT_GAP_SCAN_FAST_WINDOW,
    .options  = BT_LE_SCAN_OPT_NONE,
    .timeout  = 0,
};

static struct bt_scan_init_param scan_init = {
    .connect_if_match = true,
    .scan_param = &scan_param,
    .conn_param = NULL
};

static struct bt_scan_init_param scan_dm_init = {
    .connect_if_match = false,
    .scan_param = &scan_param,
    .conn_param = NULL
};

static struct bt_scan_manufacturer_data scan_mfg_data = {
    .data = (unsigned char*)&mfg_data,
    .data_len = sizeof(mfg_data.company_code) + sizeof(mfg_data.support_dm_code),
};

static const struct bt_le_ext_adv_cb adv_cb = {
    .scanned = adv_scanned_cb,
};

static const struct bt_data ad[] = {
    BT_DATA_BYTES(BT_DATA_FLAGS, (BT_LE_AD_GENERAL | BT_LE_AD_NO_BREDR)),
    BT_DATA(BT_DATA_NAME_COMPLETE, DEVICE_NAME, DEVICE_NAME_LEN),
};
static const struct bt_data sd[] = {
    BT_DATA(BT_DATA_MANUFACTURER_DATA, (unsigned char *)&mfg_data, sizeof(mfg_data)),
    BT_DATA_BYTES(BT_DATA_UUID128_ALL, BT_UUID_NUS_VAL),
};

static struct bt_le_adv_param adv_param_conn =
    BT_LE_ADV_PARAM_INIT(BT_LE_ADV_OPT_CONNECTABLE |
                         BT_LE_ADV_OPT_NOTIFY_SCAN_REQ,
                         BT_GAP_ADV_FAST_INT_MIN_2,
                         BT_GAP_ADV_FAST_INT_MAX_2,
                         NULL);
static struct bt_le_adv_param adv_param_noconn =
    BT_LE_ADV_PARAM_INIT(BT_LE_ADV_OPT_USE_IDENTITY |
                         BT_LE_ADV_OPT_SCANNABLE |
                         BT_LE_ADV_OPT_NOTIFY_SCAN_REQ,
                         BT_GAP_ADV_FAST_INT_MIN_2,
                         BT_GAP_ADV_FAST_INT_MAX_2,
                         NULL);

/* -------------------------------------------------------------------------- */

static void connected_cb(struct bt_conn* conn, uint8_t err)
{
    if (err)
    {
        printk("Connection failed (err %u)\n", err);
        return;
    }

    char addr[BT_ADDR_LE_STR_LEN];
    bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));
    printk("Connected to %s\n", addr);
    current_connection = bt_conn_ref(conn);

    adv_param = &adv_param_noconn;
    k_work_submit(&adv_work);
}

static void disconnected_cb(struct bt_conn* conn, uint8_t reason)
{
    char addr[BT_ADDR_LE_STR_LEN];
    bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));
    printk("Disconnected from %s (reason %u)\n", addr, reason);
    if (current_connection)
    {
        bt_conn_unref(current_connection);
        current_connection = NULL;
    }

    adv_param = NULL; // TODO
    k_work_submit(&adv_work);
}

static int adv_start(void)
{
    int err;

    if (adv)
    {
        printk("Stopping advertising\n");
        err = bt_le_ext_adv_stop(adv);
        if (err)
        {
            printk("Failed to stop extended advertising  (err %d)\n", err);
            return err;
        }
        err = bt_le_ext_adv_delete(adv);
        if (err)
        {
            printk("Failed to delete advertising set  (err %d)\n", err);
            return err;
        }
        printk("Advertising stopped\n");
        adv = NULL;
    }
    else
    {
        printk("No previous advertising\n");
    }

    printk("Starting advertising\n");
    err = bt_le_ext_adv_create(adv_param, &adv_cb, &adv);
    if (err)
    {
        printk("Failed to create advertising set (err %d)\n", err);
        return err;
    }
    err = bt_le_ext_adv_set_data(adv, ad, ARRAY_SIZE(ad), sd, ARRAY_SIZE(sd));
    if (err)
    {
        printk("Failed setting adv data (err %d)\n", err);
        return err;
    }
    err = bt_le_ext_adv_start(adv, NULL);
    if (err)
    {
        printk("Failed to start extended advertising  (err %d)\n", err);
        return err;
    }
    printk("Advertising started\n");

    return err;
}

static void adv_scanned_cb(struct bt_le_ext_adv* adv, struct bt_le_ext_adv_scanned_info* info)
{
    if (adv_param == &adv_param_noconn)
    {
        mfg_data.rng_seed = sys_cpu_to_le32(sys_rand32_get());
        int err = bt_le_ext_adv_set_data(adv, ad, ARRAY_SIZE(ad), sd, ARRAY_SIZE(sd));
        if (err)
        {
            printk("Failed setting adv data (err %d)\n", err);
            return;
        }
    }
    else
    {
        printk("Adv scanned in adv_param_conn\n");
    }
}

static int scan_dm_start(void)
{
    int err;

    bt_scan_filter_remove_all();
    bt_scan_filter_disable();

    bt_scan_init(&scan_dm_init);
    bt_scan_cb_register(&scan_dm_cb);

    err = bt_scan_filter_add(BT_SCAN_FILTER_TYPE_MANUFACTURER_DATA, &scan_mfg_data);
    if (err)
    {
        printk("Scanning filters cannot be set (err %d)\n", err);
        return err;
    }
    err = bt_scan_filter_enable(BT_SCAN_MANUFACTURER_DATA_FILTER, true);
    if (err)
    {
        printk("Filters cannot be turned on (err %d)\n", err);
        return err;
    }
    err = bt_scan_start(BT_SCAN_TYPE_SCAN_ACTIVE);
    if (err)
    {
        printk("Scanning failed to start (err %d)\n", err);
        return err;
    }
    printk("DM scanning started\n");

    struct bt_filter_status filter_status;
    bt_scan_filter_status_get(&filter_status);
    printk("UUID filter status. Enabled=%d, cnt=%d\n", (int)filter_status.uuid.enabled, filter_status.uuid.cnt);
    printk("MFG filter status. Enabled=%d, cnt=%d\n", (int)filter_status.manufacturer_data.enabled, filter_status.manufacturer_data.cnt);

    return err;
}

static void scan_filter_dm_match(struct bt_scan_device_info* device_info, struct bt_scan_filter_match* filter_match, bool connectable)
{
    static char mfg_found_addr[BT_ADDR_LE_STR_LEN];
    bt_addr_le_to_str(device_info->recv_info->addr, mfg_found_addr, sizeof(mfg_found_addr));
    if (current_connection != NULL)
    {
        printk("MFG found, connectable=%d, addr=%s\n", (int)connectable, mfg_found_addr);
        bt_addr_le_t addr;
        bt_addr_le_copy(&addr, device_info->recv_info->addr);
        bt_data_parse(device_info->adv_data, filter_data_dm_cb, &addr);
    }
    else
    {
        printk("MFG found when not connected, connectable=%d, addr=%s\n", (int)connectable, mfg_found_addr);
    }
}

static bool filter_data_dm_cb(struct bt_data* data, void* user_data)
{
    switch (data->type)
    {
    case BT_DATA_MANUFACTURER_DATA:
        // TODO
        return false;
    default:
        return true;
    }
}

static void scan_filter_nus_match(struct bt_scan_device_info* device_info, struct bt_scan_filter_match* filter_match, bool connectable)
{
    static char nus_found_addr[BT_ADDR_LE_STR_LEN];
    bt_addr_le_to_str(device_info->recv_info->addr, nus_found_addr, sizeof(nus_found_addr));
    if (current_connection != NULL)
    {
        printk("NUS filter when already connected, connectable=%d, addr=%s\n", (int)connectable, nus_found_addr);
    }
    else
    {
        printk("NUS service found, connectable=%d, addr=%s\n", (int)connectable, nus_found_addr);
    }
}

void scan_nus_connecting_error(struct bt_scan_device_info* device_info)
{
    printk("NUS service connecting error\n");
}

void scan_nus_connecting(struct bt_scan_device_info* device_info, struct bt_conn* conn)
{
    printk("NUS service connecting\n");
}

static void adv_work_cb(struct k_work* work)
{
    printk("adv_work_cb\n");
    if (adv_param == &adv_param_noconn)
    {
        int res = bt_scan_stop();
        if (res == -EALREADY)
        {
            printk("Scanning already stopped\n");
        }
        else if (res != 0)
        {
            printk("bt_scan_stop failed, res=%d\n", res);
            return;
        }
        else
        {
            printk("Scanning stopped\n");
        }
        res = adv_start();
        if (res)
        {
            printk("adv_start failed\n");
        }
        res = scan_dm_start();
        if (res)
        {
            printk("scan_dm_start failed, res=%d\n", res);
        }
    }
}

/* -------------------------------------------------------------------------- */

int main(void)
{
    mfg_data.company_code = sys_cpu_to_le16(COMPANY_CODE);
    mfg_data.support_dm_code = sys_cpu_to_le32(SUPPORT_DM_CODE);
    mfg_data.rng_seed = sys_cpu_to_le32(sys_rand32_get());

    printk("Starting Central\n");
    int err = bt_enable(NULL);
    if (err)
    {
        printk("bt_enabled failed (err %d)\n", err);
        return err;
    }
    printk("bt_enabled OK\n");

    bt_scan_init(&scan_init);
    bt_scan_cb_register(&scan_nus_cb);

    err = bt_scan_filter_add(BT_SCAN_FILTER_TYPE_UUID, BT_UUID_NUS_SERVICE);
    if (err)
    {
        printk("Scanning filters cannot be set (err %d)\n", err);
        return err;
    }
    err = bt_scan_filter_enable(BT_SCAN_UUID_FILTER, true);
    if (err)
    {
        printk("Filters cannot be turned on (err %d)\n", err);
        return err;
    }
    err = bt_scan_start(BT_SCAN_TYPE_SCAN_ACTIVE);
    if (err)
    {
        printk("Scanning failed to start (err %d)\n", err);
        return err;
    }
    printk("NUS scanning started\n");

    struct bt_filter_status filter_status;
    bt_scan_filter_status_get(&filter_status);
    printk("UUID filter status. Enabled=%d, cnt=%d\n", (int)filter_status.uuid.enabled, filter_status.uuid.cnt);
    printk("MFG filter status. Enabled=%d, cnt=%d\n", (int)filter_status.manufacturer_data.enabled, filter_status.manufacturer_data.cnt);

    while (1)
    {
        k_sleep(K_MSEC(1000));
        //printk("Central running\n");
    }

    return 0;
}

/* -------------------------------------------------------------------------- */